/***
  This file is part of systemd.

  Copyright 2010 Lennart Poettering

  systemd is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  systemd is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with systemd; If not, see <http://www.gnu.org/licenses/>.
***/

using TaoConfig;

static bool user = false;

static bool loaded;

public class ListWindow
{
  public void *handler;
}

public class JobsWindow: ListWindow
{
  private Job[] jobs;
  
  public void populate_job_model()   {
    
    int number = 0;
    jobs = new Job[0];
    
    var list = manager.list_jobs();
    
    foreach (var i in list) {
      
      Properties p = Bus.get_proxy_sync(
        user ? BusType.SESSION : BusType.SYSTEM,
        "org.freedesktop.systemd1",
        i.job_path);
      
    //  p.properties_changed.connect(on_job_changed);
      
      Job j = Bus.get_proxy_sync(
        user ? BusType.SESSION : BusType.SYSTEM,
        "org.freedesktop.systemd1",
        i.job_path);
      
     /* Tao.add_handler(handler, "/list/items/" + number.to_string() + "/data/Unit State", null, null);
      
      Tao.add_handler(handler, "/list/items/" + number.to_string() + "/data/Active State", null, null);
      
      Tao.add_handler(handler, "/list/items/" + number.to_string() + "/data/Load State", null, null);
      
      Tao.add_handler(handler, "/list/items/" + number.to_string() + "/data/Unit", null, null);
      
      Tao.set_hint(this.handler, "/list/items/" + number.to_string() + "/data/Unit State", '\0', i.sub_state);
      
      Tao.set_hint(handler, "/list/items/" + number.to_string() + "/data/Active State", '\0', i.active_state);
      
      Tao.set_hint(handler, "/list/items/" + number.to_string() + "/data/Load State", '\0', i.load_state);
      
      Tao.set_hint(this.handler, "/list/items/" + number.to_string() + "/data/Unit", '\0', i.description);
      
      Tao.add_handler(handler, "/list/items/" + number.to_string() + "/selected",(void*) select_unit, u);*/
      
      jobs += j;
      Tao.run_event_iterration();
      
      ++number;
    }
    }
    
}

public class MainWindow
{
  public void *handler;
  
  public MainWindow() {
    
    var title = user ? "User session service management" : "System service management";
    
    handler = Tao.new_window("/desktop/dialogs/selection");
    
    Tao.set_hint(handler, null, 0, title);
    
    manager = Bus.get_proxy_sync(
      user ? BusType.SESSION : BusType.SYSTEM,
      "org.freedesktop.systemd1",
      "/org/freedesktop/systemd1");
    
   // manager.job_removed.connect(on_job_removed);
    
    Tao.add_handler(handler, "/:abort", (void*)app_exit, (void*)0);
    
    Tao.add_handler(handler, "/list/columns/configure", null, null);
    Tao.add_handler(handler, "/list/items/units/data/configure", null, null);
    Tao.add_handler(handler, "/list/items/jobs/data/configure", null, null);
    
    Tao.add_handler(handler, "/list/items/units/selected", (void*) start_uwindow, null);
    Tao.add_handler(handler, "/list/items/jobs/selected", (void*) start_jwindow, null);
    
    Tao.set_hint(handler, "/list/items/units/data/configure", '\0', "Units");
    Tao.set_hint(handler, "/list/items/jobs/data/configure", '\0', "Jobs");
    
  }
  
}

void close_window(void *mess, void *mess2)
{
  void **windptr = (void**) mess2;
  Tao.release_window(((ListWindow*)(*windptr))->handler);
  *windptr = null;
  
  mwindow = new MainWindow();
}

void start_uwindow(void *mess, void *mess2)
{
  
  if (uwindow != null) {
  
    return;
  }
  void *splashw = Tao.new_window("/desktop/dialogs/message_dialog");
  Tao.add_handler(splashw, "/:abort", (void*)app_exit, (void*)0);
  
  
  Tao.add_handler(splashw, "/message", null, null);
  Tao.set_hint(splashw, "/message", '\0', "Loading");
  uwindow = new UnitsWindow();
  
  Tao.set_modal(uwindow.handler, "/splash", splashw, 1);
  
  uwindow.populate_unit_model();
  
  Tao.release_window(splashw);
  
  Tao.add_handler(uwindow.handler, "/:abort", (void*)close_window, &uwindow);
  
  void *switchw;
  
  switchw = Tao.new_window("/desktop/dialogs/selection");
  
  Tao.attach_window(uwindow.handler, "/objects-related", switchw);
  
  initialize_filter_window(switchw);
  
  Tao.release_window(mwindow.handler);
  
}

void start_jwindow(void *mess, void *mess2)
{
}

public class UnitsWindow: ListWindow
{        
        public int current_type_id;
        public Unit *current_unit;
        //public Unit_Map[] umap;
        private Unit[] units;
        
        private string unit_information;
        
        public UnitsWindow() /*  */{
          
          handler = Tao.new_window("/Objects/MultipleObjects");
          
          initialize_columns_of_units(handler);
          
          Tao.add_handler(handler, "/actions/Start", (void*)on_start, null);
          Tao.add_handler(handler, "/actions/Stop", (void*) on_stop, null);
          Tao.add_handler(handler, "/actions/Restart", (void*) on_restart, null);
          Tao.add_handler(handler, "/actions/Reload", (void*) on_reload, null);
          
                /*title = user ? "systemd User Service Manager" : "systemd System Manager";
                clear_unit();
                clear_job();
                populate_unit_model();
                populate_job_model();
                */
        }

        private string get_selected_type()
        {
            switch (current_type_id) {
            
              default:
              case 0:
                return "";
              case 1:
                return ".target";
              case 2:
                return ".slice";
              case 3:
                return ".scope";
              case 4:
                return ".service";
              case 5:
                return ".device";
              case 6:
                return ".mount";
              case 7:
                return ".automount";
              case 8:
                return ".swap";
              case 9:
                return ".socket";
              case 10:
                return ".path";
              case 11:
                return ".timer";
              case 12:
                return ".snapshot";
            }
        }
        
        public void populate_unit_model() /* */ {
                /*unit_model.clear();
*/
                int number = 0;
                var list = manager.list_units();
                string suffix = get_selected_type();

                
                Tao.delete_handlers(handler, "/list/items/");
//                umap = [];
                
                
                units = new Unit[0];
                
                foreach (var i in list) {
 
                        
    
                        if (i.id.has_suffix(suffix)) {
                        
                          Properties p = Bus.get_proxy_sync(
                            user ? BusType.SESSION : BusType.SYSTEM,
                            "org.freedesktop.systemd1",
                            i.unit_path);
                          
                          //p.properties_changed.connect(on_unit_changed);
                          
                          Unit u = Bus.get_proxy_sync(
                            user ? BusType.SESSION : BusType.SYSTEM,
                            "org.freedesktop.systemd1",
                            i.unit_path);
                          
                        Tao.add_handler(handler, "/list/items/" + number.to_string() + "/data/Unit State", null, null);
                       
                        Tao.add_handler(handler, "/list/items/" + number.to_string() + "/data/Active State", null, null);
    
                        Tao.add_handler(handler, "/list/items/" + number.to_string() + "/data/Load State", null, null);
                        
                        Tao.add_handler(handler, "/list/items/" + number.to_string() + "/data/Unit", null, null);
                        
                        Tao.set_hint(this.handler, "/list/items/" + number.to_string() + "/data/Unit State", '\0', i.sub_state);
                        
                        Tao.set_hint(handler, "/list/items/" + number.to_string() + "/data/Active State", '\0', i.active_state);
                        
                        Tao.set_hint(handler, "/list/items/" + number.to_string() + "/data/Load State", '\0', i.load_state);
                        
                        Tao.set_hint(this.handler, "/list/items/" + number.to_string() + "/data/Unit", '\0', i.description);
                        
                        Tao.add_handler(handler, "/list/items/" + number.to_string() + "/selected",(void*) select_unit, u);
                        
                        units += u;
                        Tao.run_event_iterration();
                        
                        }
                        ++number;
                }
                
        }

        public void unit_changed() {
               /* Unit u = get_current_unit();

                if (u == null)
                        clear_unit();
                else
                        show_unit(u);
                */
        }

        public string make_dependency_string(string? prefix, string word, string[] dependencies) {
              
                bool first = true;
                string r ="";
                
                foreach (string i in dependencies) {
                
                     if (r != "")
                       r += first ? "\n" : ",";
                     
                     if (first)  {
                        
                       r += word + ":";
                       first = false;
                     }
                     
                     r += i;
            
                }
                
                return r;
        }

       public void show_unit(Unit unit) {
            
            string[]
            requires = unit.requires,
            requires_overridable = unit.requires_overridable,
            requisite = unit.requisite,
            requisite_overridable = unit.requisite_overridable,
            wants = unit.wants,
            required_by = unit.required_by,
            required_by_overridable = unit.required_by_overridable,
            wanted_by = unit.wanted_by,
            conflicts = unit.conflicts,
            before = unit.before,
            after = unit.after;
            
            unit_information = 
            make_dependency_string(
              make_dependency_string(
                make_dependency_string(
                  make_dependency_string(
                    make_dependency_string(
                      make_dependency_string(
                        make_dependency_string(
                          make_dependency_string(
                            make_dependency_string(
                              make_dependency_string(
                                make_dependency_string(null,
                                                       "requires", requires),
                                                     "overridable requires", requires_overridable),
                                                   "requisite", requisite),
                                                 "overridable requisite", requisite_overridable),
                                               "wants", wants),
                                             "conflicts", conflicts),
                                           "required by", required_by),
                                         "overridable required by", required_by_overridable),
                                       "wanted by", wanted_by),
                                     "after", after),
                                   "before", before);
                                   
            /* It looks like error in libgreattao - this line is necessary(unfortunally)*/
              Tao.add_handler(handler, "/information", null, null);
              Tao.add_handler(handler, "/information/custom/text/dependency/value", null, null);
              Tao.set_hint(handler, "/information/custom/text/dependency/value", 0, unit_information);
              Tao.add_handler(handler, "/information/custom/text/dependency/name", null, null);
              Tao.set_hint(handler, "/information/custom/text/dependency/name", 0, "Dependencies");
              
              Tao.add_handler(handler, "/information/custom/text/description/name", null, null);
              Tao.set_hint(handler, "/information/custom/text/description/name", 1, "Description");
              
              Tao.add_handler(handler, "/information/custom/text/description/value", null, null);
              Tao.set_hint(handler, "/information/custom/text/description/value", 0, unit.description);
              
              Tao.add_handler(handler, "/information/custom/text/load state/name", null, null);
              Tao.set_hint(handler, "/information/custom/text/load state/name", 1, "Load state");
              
              Tao.add_handler(handler, "/information/custom/text/load state/value", null, null);
              Tao.set_hint(handler, "/information/custom/text/load state/value", 0, unit.load_state);
       }

        public void job_changed() {
          /*      
          Job j = get_current_job();

                if (j == null)
                        clear_job();
                else
                        show_job(j);
                */
        }

        public void clear_job() {
        }

        
        public void update_unit(int number, Unit u)
        {
          Tao.add_handler(handler, "/list/items/" + number.to_string() + "/data/Unit State", null, null);
          
          Tao.add_handler(handler, "/list/items/" + number.to_string() + "/data/Active State", null, null);
          
          Tao.add_handler(handler, "/list/items/" + number.to_string() + "/data/Load State", null, null);
          
          Tao.add_handler(handler, "/list/items/" + number.to_string() + "/data/Unit", null, null);
          
          Tao.set_hint(this.handler, "/list/items/" + number.to_string() + "/data/Unit State", '\0', u.sub_state);
          
          Tao.set_hint(handler, "/list/items/" + number.to_string() + "/data/Active State", '\0', u.active_state);
          
          Tao.set_hint(handler, "/list/items/" + number.to_string() + "/data/Load State", '\0', u.load_state);
          
          Tao.set_hint(this.handler, "/list/items/" + number.to_string() + "/data/Unit", '\0', u.description);
          
        }

        public void on_unit_new(string id, ObjectPath path) {
               try {

                        Properties p = Bus.get_proxy_sync(
                                        user ? BusType.SESSION : BusType.SYSTEM,
                                        "org.freedesktop.systemd1",
                                        path);

                      //  p.properties_changed.connect(on_unit_changed);

                        Unit u = Bus.get_proxy_sync(
                                        user ? BusType.SESSION : BusType.SYSTEM,
                                        "org.freedesktop.systemd1",
                                        path);

                } catch (Error e) {
                        show_error(e);
                }
                
        }

        public void on_job_new(uint32 id, ObjectPath path) {

        }

        public void on_unit_removed(string id, ObjectPath path) {
                
          
              Tao.delete_handlers(handler, "/list/items" + id.to_string());
        }

        public void on_job_removed(uint32 id, ObjectPath path, string res) {
                
        }


        public void unit_type_changed(void *mess1, void *number) {
          
          int type = (int) number;
          
          // TODO: Complete this.
    
        }

        public void on_server_reload() {
              /*  try {
                        manager.reload();
                } catch (Error e) {
                        show_error(e);
                }*/
        }

        public void on_server_snapshot() {
               /* try {
                        manager.create_snapshot();

                        if (unit_type_combo_box.get_active() != 0)
                                unit_type_combo_box.set_active(12);
                } catch (Error e) {
                        show_error(e);
                }*/
        }
}


void show_error(Error e) {
  string title, message;
  if (e is DBusError && DBusError.is_remote_error(e)) {
    string remote = DBusError.get_remote_error(e);
    message = strerror(dbus_error_to_errno(remote)) + ":" + remote;
  } else {
    message = e.message;
  }
  TaoLog.custom("<UNKOWN>", "main", "Error occured while sending message to bus", 
                message,"systemadm.vala", -1, 1|2);
  
}

UnitsWindow uwindow;
JobsWindow jwindow;
MainWindow mwindow;
public Manager manager;

int dbus_error_to_errno(string error) {
		switch (error) {
		case "org.freedesktop.DBus.Error.Failed":
				return Posix.EACCES;
		case "org.freedesktop.DBus.Error.NoMemory":
				return Posix.ENOMEM;
		case "org.freedesktop.DBus.Error.ServiceUnknown":
				return Posix.EHOSTUNREACH;
		case "org.freedesktop.DBus.Error.NameHasNoOwner":
				return Posix.ENXIO;
		case "org.freedesktop.DBus.Error.NoReply":
				return Posix.ETIMEDOUT;
		case "org.freedesktop.DBus.Error.IOError":
				return Posix.EIO;
		case "org.freedesktop.DBus.Error.BadAddress":
				return Posix.EADDRNOTAVAIL;
		case "org.freedesktop.DBus.Error.NotSupported":
				return Posix.ENOTSUP;
		case "org.freedesktop.DBus.Error.LimitsExceeded":
				return Posix.ENOBUFS;
		case "org.freedesktop.DBus.Error.AccessDenied":
				return Posix.EACCES;
		case "org.freedesktop.DBus.Error.AuthFailed":
				return Posix.EACCES;
		case "org.freedesktop.DBus.Error.NoServer":
				/* Errno missing from Posix */
				/* return Posix.EHOSTDOWN; */
				return Posix.EIO;
		case "org.freedesktop.DBus.Error.Timeout":
				return Posix.ETIMEDOUT;
		case "org.freedesktop.DBus.Error.NoNetwork":
				/* Errno missing from Posix */
				/* return Posix.ENONET; */
				return Posix.EIO;
		case "org.freedesktop.DBus.Error.AddressInUse":
				return Posix.EADDRINUSE;
		case "org.freedesktop.DBus.Error.Disconnected":
				return Posix.ECONNRESET;
		case "org.freedesktop.DBus.Error.InvalidArgs":
				return Posix.EINVAL;
		case "org.freedesktop.DBus.Error.FileNotFound":
				return Posix.ENOENT;
		case "org.freedesktop.DBus.Error.FileExists":
				return Posix.EEXIST;
		case "org.freedesktop.DBus.Error.UnknownMethod":
		case "org.freedesktop.DBus.Error.UnknownObject":
		case "org.freedesktop.DBus.Error.UnknownInterface":
		case "org.freedesktop.DBus.Error.UnknownProperty":
				/* Errno missing from Posix */
				/* return Posix.EBADR; */
				return Posix.EIO;
		case "org.freedesktop.DBus.Error.PropertyReadOnly":
				return Posix.EROFS;
		case "org.freedesktop.DBus.Error.UnixProcessIdUnknown":
				return Posix.ESRCH;
		case "org.freedesktop.DBus.Error.InvalidSignature":
				return Posix.EINVAL;
		case "org.freedesktop.DBus.Error.InconsistentMessage":
				return Posix.EBADMSG;
		case "org.freedesktop.DBus.Error.TimedOut":
				return Posix.ETIMEDOUT;
		case "org.freedesktop.DBus.Error.MatchRuleInvalid":
		case "org.freedesktop.DBus.Error.InvalidFileContent":
				return Posix.EINVAL;
		case "org.freedesktop.DBus.Error.MatchRuleNotFound":
				return Posix.ENOENT;
		case "org.freedesktop.DBus.Error.SELinuxSecurityContextUnknown":
				return Posix.ESRCH;
		case "org.freedesktop.DBus.Error.ObjectPathInUse":
				return Posix.EBUSY;
		default:
				return Posix.EIO;
		}
}

void get_polkit_access(string path) {

  var authority = Polkit.Authority.get_sync(null);
  var subject = new Polkit.UnixProcess(Posix.getpid());
  var auth  =  Polkit.Authority.get();
  var is_ok = auth.check_authorization_sync(subject, path, null, Polkit.CheckAuthorizationFlags.ALLOW_USER_INTERACTION, null);
  
  if (is_ok.get_is_authorized()) {
  
    stdout.printf("OK!\n");
  }
}

void app_exit(void *m1, void *m2)
{
  int result = (int) m2;
  
  Tao.close();
  
  Posix.exit(result);
}

void select_unit(void *mess, void *unit_)
{
  Unit *unit = (Unit*) unit_;
  
  uwindow.current_unit = unit;
  uwindow.show_unit(unit);
}

void on_stop() {
  if (uwindow.current_unit == null) {
    
    return;
  }
  
  try {
    uwindow.current_unit->stop();
  }
  catch (DBusError e) {
    
    if (DBusError.is_remote_error(e)) {
      string remote = DBusError.get_remote_error(e);
      
      if (dbus_error_to_errno(remote) == Posix.EACCES) {
        
        // get_polkit_access("org.freedesktop.systemd1.reload-daemon");
        get_polkit_access("org.freedesktop.systemd1.manage-units");
        try {
          uwindow.current_unit->stop();
        }
        catch (Error e) {
          
          show_error(e);
        }
      }
    }
    else {
      
      show_error(e);
    }
    
  }
  catch (Error e) {
    
    show_error(e);
  }
}

void on_reload() {
  if (uwindow.current_unit == null) {
    
    return;
  }
  
  try {
    uwindow.current_unit->reload();
  }
  catch (DBusError e) {
    
    if (DBusError.is_remote_error(e)) {
      string remote = DBusError.get_remote_error(e);
      
      if (dbus_error_to_errno(remote) == Posix.EACCES) {
        
        // get_polkit_access("org.freedesktop.systemd1.reload-daemon");
        get_polkit_access("org.freedesktop.systemd1.manage-units");
        try {
          uwindow.current_unit->reload();
        }
        catch (Error e) {
          
          show_error(e);
        }
      }
    }
    else {
      
      show_error(e);
    }
    
  }
  catch (Error e) {
    
    show_error(e);
  }
    
}

void on_restart() {
  if (uwindow.current_unit == null) {
    
    return;
  }
  
  try {
    uwindow.current_unit->restart();
  }
  catch (DBusError e) {
    
    if (DBusError.is_remote_error(e)) {
      string remote = DBusError.get_remote_error(e);
      
      if (dbus_error_to_errno(remote) == Posix.EACCES) {
        
        // get_polkit_access("org.freedesktop.systemd1.reload-daemon");
        get_polkit_access("org.freedesktop.systemd1.manage-units");
        try {
          uwindow.current_unit->restart();
        }
        catch (Error e) {
          
          show_error(e);
        }
      }
    }
    else {
      
      show_error(e);
    }
    
  }
  catch (Error e) {
    
    show_error(e);
  }
    
}

void on_cancel() {
  /*      
   * Job j = get_current_job();
   * 
   *       if (j == null)
   *               return;
   * 
   *       try {
   *               j.cancel();
} catch (Error e) {
show_error(e);
}
*/
}

void on_start() {
  if (uwindow.current_unit == null) {
  
    return;
  }
  
  try {
    uwindow.current_unit->start("replace");
  }
  catch (DBusError e) {
  
    if (DBusError.is_remote_error(e)) {
      string remote = DBusError.get_remote_error(e);
      
      if (dbus_error_to_errno(remote) == Posix.EACCES) {
      
       // get_polkit_access("org.freedesktop.systemd1.reload-daemon");
        get_polkit_access("org.freedesktop.systemd1.manage-units");
        try {
          uwindow.current_unit->start("replace");
          
        }
        catch (Error e) {
          show_error(e);
        }
      }
      else {
        
        show_error(e);
      }
    }
    else {
      
      show_error(e);
    }
      
  }
  catch (Error e) {
    
    show_error(e);
  }
}

char initialize_columns_of_units(void *window)
{
  Tao.add_handler(window, "/list/columns/Load State", null, null);
  Tao.add_handler(window, "/list/columns/Active State", null, null);
  Tao.add_handler(window, "/list/columns/Unit State", null, null);
  Tao.add_handler(window, "/list/columns/Unit", null, null);
  
  return 1;
}

void set_unit_filter(void *mess1, void *number)
{
  int id = (int) number;
  
  uwindow.current_type_id = id;
  
  uwindow.populate_unit_model();
}

void initialize_filter_window(void *window)
{
  Tao.add_handler(window, "/list/columns/Unit Types", null, null);
  
  Tao.add_handler(window, "/list/items/0/data/Unit Types", null, null);
  Tao.add_handler(window, "/list/items/0/selected", (void*)set_unit_filter, (void*) 0);
  Tao.set_hint(window, "/list/items/0/data/Unit Types", 0, "All unit types");

  
  Tao.add_handler(window, "/list/items/1/data/Unit Types", null, null);
  Tao.add_handler(window, "/list/items/1/selected", (void*)set_unit_filter, (void*) 1);
  Tao.set_hint(window, "/list/items/1/data/Unit Types", 0, "Targets");
  
  Tao.add_handler(window, "/list/items/2/data/Unit Types", null, null);
  Tao.add_handler(window, "/list/items/2/selected", (void*)set_unit_filter, (void*) 2);
  Tao.set_hint(window, "/list/items/2/data/Unit Types", 0, "Slices");
  
  Tao.add_handler(window, "/list/items/3/data/Unit Types", null, null);
  Tao.add_handler(window, "/list/items/3/selected", (void*)set_unit_filter, (void*) 3);
  Tao.set_hint(window, "/list/items/3/data/Unit Types", 0, "Scopes");
  
  Tao.add_handler(window, "/list/items/4/data/Unit Types", null, null);
  Tao.add_handler(window, "/list/items/4/selected", (void*)set_unit_filter, (void*) 4);
  Tao.set_hint(window, "/list/items/4/data/Unit Types", 0, "Services");
  
  Tao.add_handler(window, "/list/items/5/data/Unit Types", null, null);
  Tao.add_handler(window, "/list/items/5/selected", (void*)set_unit_filter, (void*) 5);
  Tao.set_hint(window, "/list/items/5/data/Unit Types", 0, "Devices");
  
  Tao.add_handler(window, "/list/items/6/data/Unit Types", null, null);
  Tao.add_handler(window, "/list/items/6/selected", (void*)set_unit_filter, (void*) 6);
  Tao.set_hint(window, "/list/items/6/data/Unit Types", 0, "Mounts");
  
  Tao.add_handler(window, "/list/items/7/data/Unit Types", null, null);
  Tao.add_handler(window, "/list/items/7/selected", (void*)set_unit_filter, (void*) 7);
  Tao.set_hint(window, "/list/items/7/data/Unit Types", 0, "Automounts");
  
  Tao.add_handler(window, "/list/items/8/data/Unit Types", null, null);
  Tao.add_handler(window, "/list/items/8/selected", (void*)set_unit_filter, (void*) 8);
  Tao.set_hint(window, "/list/items/8/data/Unit Types", 0, "Swaps");
  
  Tao.add_handler(window, "/list/items/9/data/Unit Types", null, null);
  Tao.add_handler(window, "/list/items/9/selected", (void*)set_unit_filter, (void*) 9);
  Tao.set_hint(window, "/list/items/9/data/Unit Types", 0, "Sockets");
  
  Tao.add_handler(window, "/list/items/10/data/Unit Types", null, null);
  Tao.add_handler(window, "/list/items/10/selected", (void*)set_unit_filter, (void*) 10);
  Tao.set_hint(window, "/list/items/10/data/Unit Types", 0, "Paths");
  
  Tao.add_handler(window, "/list/items/11/data/Unit Types", null, null);
  Tao.add_handler(window, "/list/items/11/selected", (void*)set_unit_filter, (void*) 11);
  Tao.set_hint(window, "/list/items/11/data/Unit Types", 0, "Timers");
  
  Tao.add_handler(window, "/list/items/12/data/Unit Types", null, null);
  Tao.add_handler(window, "/list/items/12/selected", (void*)set_unit_filter, (void*) 12);
  Tao.set_hint(window, "/list/items/12/data/Unit Types", 0, "Snapshots");
  
}

tao_app_config_option[] app_config;
tao_app_config_result[] conf_result;

int main(string[] args) {

        app_config += tao_app_config_option();
        conf_result+= tao_app_config_result();
        
        app_config[0].option_name = "scope";
        app_config[0].visible_name = "scope";
        app_config[0].visible_description = "scope";
        app_config[0].type = TaoConfig.TAO_TYPE_STRING;
        app_config[0].short_option = '\0';
        app_config[0].flags = TaoConfig.TAO_CONFIG_OPTION_ARG_TAO_DECORATOR | 
        TaoConfig.TAO_CONFIG_OPTION_ARG_LONG_DECORATOR;
        
        Tao.initialize("systemadm", "", &args.length, args);
  
        TaoConfig.tao_get_app_configuration(&app_config[0], &conf_result[0], 1, TaoConfig.TAO_APP_CONFIG_FROM_LINE);
        
        if (conf_result[0].set_ == 1 && conf_result[0].error == 0) {
        
          stdout.printf(((string)conf_result[0].result_)+"\n");
          if (((string)conf_result[0].result_) == "user") {
          
            user = true;
          }
          else if (((string)conf_result[0].result_) == "system") {
          
            user = false;
          }
          else {
          
            TaoLog.custom("<UNKOWN>", "main", "Error in parameters", 
                          "To scope parameter can been assigned only user or system","systemadm.vala", -1, 1|2);
            Posix.exit(1);
          }
        }
        
        manager.subscribe();
        
        manager.unit_new.connect(uwindow.on_unit_new);
        manager.job_new.connect(uwindow.on_job_new);
        manager.unit_removed.connect(uwindow.on_unit_removed);
        
        mwindow = new MainWindow();
        
        Tao.handle_events();
        return 0;
}
